#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

// types ////////////////////////////////////////////////////////////////////

typedef struct{
   float transform[4][4]; // from world-space to camera-space
   float inverse[4][4]; // from camera-space to world-space
   float camera_x_max; // camera-space x clipped at +/- camera_x_max.
                       // camera-space y clipped at -1 and 1
                       // camera-space z=0 is at the camera, z=1 is at the center of view.
} ViewSettings;

typedef struct FragmentList FragmentList;
struct FragmentList{
  float n, nmax;
  float *rgbaz;
  struct FragmentList *next;
};

typedef struct{
   int m, n;
   FragmentList **fragments;
} FragmentBuffer;

typedef struct{
   int m, n;
   float *r, *g, *b, *a, *z;
} ImageRGBAZ;

typedef struct{
   int n;
   float *x, *rgb, *radius, *opacity;
} Particles;

// main prototypes //////////////////////////////////////////////////////////

void read_view(const char *filename, ViewSettings *view, FragmentBuffer *fb, ImageRGBAZ *img, float *exposure);
void read_particles(const char *filename, Particles *pts);
void write_image(const char *filename, float exposure, const ImageRGBAZ *img);
void fragment_particles(const Particles *pts, const ViewSettings *view, FragmentBuffer *fb);
void composite_fragments(FragmentBuffer *fb, ImageRGBAZ *img);

// code /////////////////////////////////////////////////////////////////////

int main(int argc, const char **argv)
{
   ViewSettings view;
   FragmentBuffer fb;
   ImageRGBAZ img;
   float exposure;
   Particles pts;

   if(argc<4){
      fprintf(stderr,"usage: renderparticles viewfile particlefile imagefile\n");
      return 1;
   }
   printf("Reading data...\n");
   read_view(argv[1], &view, &fb, &img, &exposure);
   read_particles(argv[2], &pts);
   printf("Scan converting...\n");
   fragment_particles(&pts, &view, &fb);
   printf("Compositing...\n");
   composite_fragments(&fb, &img);
   printf("Writing image...\n");
   write_image(argv[3], exposure, &img);

   return 0;
}

// View stuff ///////////////////////////////////////////////////////////////////////////////////

// composition multiplies on the right (like OpenGL)
void compose_perspective(float transform[4][4], float inverse[4][4], float characteristic_depth, float characteristic_height);
void compose_rotation(float transform[4][4], float inverse[4][4], int axis, float degrees);
void compose_lookat(float transform[4][4], float inverse[4][4], const float cam[3], const float centre[3]);
void compose_translation(float transform[4][4], float inverse[4][4], float x, float y, float z);
inline float sqr(float x) {return x*x;}
inline float dist(const float u[3], const float v[3]) {return sqrt(sqr(u[0]-v[0])+sqr(u[1]-v[1])+sqr(u[2]-v[2]));}

void read_view(const char *filename, ViewSettings *view, FragmentBuffer *fb, ImageRGBAZ *img, float *exposure)
{
   FILE *viewfile=fopen(filename, "rt");
   int i, j;
   float cam[3], centre[3], centreheight, camroll;

   if(!viewfile){
      fprintf(stderr, "Could not open %s for reading camera settings\n", filename);
      exit(1);
   }
   // default values for view
   for(i=0; i<4; ++i)
      for(j=0; j<4; ++j){
         view->transform[i][j]=0;
         view->inverse[i][j]=0;
      }
   for(i=0; i<4; ++i){
      view->transform[i][i]=1;
      view->inverse[i][i]=1;
   }
   view->camera_x_max=4./3.;
   // default values for fragment buffer
   fb->m=640;
   fb->n=480;
   fb->fragments=0;
   // default values for image
   img->m=640;
   img->n=480;
   img->r=img->g=img->b=img->a=img->z=0;
   // default exposure
   *exposure=1;

   // viewfile - camera x y z, center of vision x y z, height at center, angles rotated off vertical,
   //            exposure (what to scale final rgb by to get back to [0,1] hopefully)
   //            image width and height
   fscanf(viewfile, "%f %f %f %f %f %f %f %f", cam, cam+1, cam+2, centre, centre+1, centre+2, &centreheight, &camroll);
   fscanf(viewfile, "%f %d %d", exposure, &fb->m, &fb->n);
   fclose(viewfile);

   // code to get the camera transformation as a matrix
   compose_perspective(view->transform, view->inverse, dist(cam, centre), centreheight);
   compose_rotation(view->transform, view->inverse, 2, camroll);
   compose_lookat(view->transform, view->inverse, cam, centre);
   // work out image (viewport) stuff
   view->camera_x_max=fb->m/(float)fb->n;
   fb->fragments=(FragmentList**)calloc(fb->m*fb->n, sizeof(FragmentList*));
   img->m=fb->m;
   img->n=fb->n;
   img->r=(float*)calloc(5*img->m*img->n, sizeof(float));
   img->g=img->r+img->m*img->n;
   img->b=img->g+img->m*img->n;
   img->a=img->b+img->m*img->n;
   img->z=img->a+img->m*img->n;
}

// Does a perspective transform that maps world: (x, y, z) to camera: (-(x/h)/(z/d), -(y/h)/(z/d), 1+d/z)
// So a point in the world at (0,h,-d) is mapped to (0,1,0), and squares facing the camera project to squares (not rectangles).
// The camera is at the origin, looking down the negative z-axis, with x increasing to the right, y increasing upwards.
void compose_perspective(float transform[4][4], float inverse[4][4], float characteristic_depth, float characteristic_height)
{
   int i;
   float t;
   for(i=0; i<4; ++i){
      transform[i][0]*=characteristic_depth/characteristic_height;
      transform[i][1]*=characteristic_depth/characteristic_height;
      t=transform[i][2];
      transform[i][2]=-t-transform[i][3];
      transform[i][3]=-characteristic_depth*t;
      inverse[0][i]*=characteristic_height/characteristic_depth;
      inverse[1][i]*=characteristic_height/characteristic_depth;
      t=inverse[2][i];
      inverse[2][i]=-inverse[3][i];
      inverse[3][i]=(inverse[3][i]-t)/characteristic_depth;
   }
}

void compose_rotation(float transform[4][4], float inverse[4][4], int axis, float degrees)
{
   float c=cos(degrees*M_PI/180), s=sin(degrees*M_PI/180), t;
   int i, a1=(axis+1)%3, a2=(axis+2)%3;
   for(i=0; i<4; ++i){
      t=transform[i][a1];
      transform[i][a1]=c*t+s*transform[i][a2];
      transform[i][a2]=c*transform[i][a2]-s*t;
      t=inverse[a1][i];
      inverse[a1][i]=c*t+s*inverse[a2][i];
      inverse[a2][i]=c*inverse[a2][i]-s*t;
   }
}

// rotation and translation s.t. cam becomes the origin, centre is on the negative z axis,
// and transformed x axis is parallel to the original x-y plane.
void compose_lookat(float transform[4][4], float inverse[4][4], const float cam[3], const float centre[3])
{
   float R[3][3], t[3];
   int i, j;
   if(cam[0]==centre[0] && cam[1]==centre[1]){
      float s=(cam[0]>=centre[0]) ? 1 : -1;
      R[0][0]=1; R[0][1]=0; R[0][2]=0;
      R[1][0]=0; R[1][1]=s; R[1][2]=0;
      R[2][0]=0; R[2][1]=0; R[2][2]=s;
   }else{
      float mag;
      R[2][0]=cam[0]-centre[0]; R[2][1]=cam[1]-centre[1]; R[2][2]=cam[2]-centre[2];
      mag=sqrt(sqr(R[2][0])+sqr(R[2][1])+sqr(R[2][2]));
      R[2][0]/=mag; R[2][1]/=mag; R[2][2]/=mag;
      R[0][0]=-R[2][1]; R[0][1]=R[2][0]; R[0][2]=0;
      mag=sqrt(sqr(R[0][0])+sqr(R[0][1]));
      R[0][0]/=mag; R[0][1]/=mag;
      R[1][0]=-R[2][2]*R[0][1];
      R[1][1]=R[2][2]*R[0][0];
      R[1][2]=R[2][0]*R[0][1]-R[2][1]*R[0][0];
   }
   for(i=0; i<4; ++i){
      // multiply by R on the right of T
      for(j=0; j<3; ++j)
         t[j]=transform[i][j];
      for(j=0; j<3; ++j)
         transform[i][j]=t[0]*R[0][j]+t[1]*R[1][j]+t[2]*R[2][j];
      // now multiply by R^T on the left of T^(-1)
      for(j=0; j<3; ++j)
         t[j]=inverse[j][i];
      for(j=0; j<3; ++j)
         inverse[j][i]=R[0][j]*t[0]+R[1][j]*t[1]+R[2][j]*t[2];
   }
   compose_translation(transform, inverse, -cam[0], -cam[1], -cam[2]);
}

void compose_translation(float transform[4][4], float inverse[4][4], float x, float y, float z)
{
   int i;
   for(i=0; i<4; ++i){
      transform[i][3]+=x*transform[i][0]+y*transform[i][1]+z*transform[i][2];
      inverse[0][i]-=x*inverse[3][i];
      inverse[1][i]-=y*inverse[3][i];
      inverse[2][i]-=z*inverse[3][i];
   }
}

// Read particles ///////////////////////////////////////////////////////////////////////////////

#define LINESIZE 1024
void read_particles(const char *filename, Particles *pts)
{
   FILE *ptsfile=fopen(filename, "rt");
   char line[LINESIZE];
   int i;

   if(!ptsfile){
      fprintf(stderr, "Could not open %s for reading particles\n", filename);
      exit(1);
   }
   // particlefile - #particles, newline, particle_1: x y z r g b radius [opacity], newline, ...
   fscanf(ptsfile, "%d", &pts->n);
   pts->x=(float*)malloc(8*pts->n*sizeof(float));
   pts->rgb=pts->x+3*pts->n;
   pts->radius=pts->rgb+3*pts->n;
   pts->opacity=pts->radius+pts->n;
   for(i=0; i<pts->n; ++i){
      fgets(line, LINESIZE, ptsfile);
      pts->opacity[i]=1; // default opacity
      sscanf(line, "%f %f %f %f %f %f %f %f", pts->x+3*i, pts->x+3*i+1, pts->x+3*i+2,
                                              pts->rgb+3*i, pts->rgb+3*i+1, pts->rgb+3*i+2,
                                              pts->radius+i, pts->opacity+i);
   }
   fclose(ptsfile);
}

// Write image //////////////////////////////////////////////////////////////////////////////////

void write_bigendian_short(FILE *fp, short int x)
{
   char buf[2];
   buf[0]=x>>8;
   buf[1]=x&0xFF;
   fwrite(buf, 1, 2, fp);
}

void write_bigendian_word(FILE *fp, int x)
{
   char buf[4];
   buf[0]=x>>24;
   buf[1]=(x>>16)&0xFF;
   buf[2]=(x>>8)&0xFF;
   buf[3]=x&0xFF;
   fwrite(buf, 1, 4, fp);
}

void write_zeros(FILE *fp, int num_zero_bytes)
{
   int i;
   for(i=0; i<num_zero_bytes; ++i)
      fputc(0, fp);
}

void write_channel(FILE *fp, float *channel, int size, float exposure)
{
   int i;
   for(i=0; i<size; ++i){
      unsigned char vc;
      float vf=256*(channel[i]*exposure);
      if(vf>255) vf=255; else if(vf<0) vf=0;
      vc=(unsigned char)vf;
      fputc(vc, fp);
   }
}

void write_image(const char *filename, float exposure, const ImageRGBAZ *img)
{
   FILE *out=fopen(filename, "wb");

   if(!out){
      fprintf(stderr, "Could not open %s for writing image\n", filename);
      exit(1);
   }
   write_bigendian_short(out, 474); // magic number
   fputc(0, out); // verbatim format
   fputc(1, out); // 1 byte per pixel channel
   write_bigendian_short(out, 3); // 3 dimensions: x, y, channels
   write_bigendian_short(out, img->m);
   write_bigendian_short(out, img->n);
   write_bigendian_short(out, 3); // 3 channels (RGB)
   write_bigendian_word(out, 0); // min pixel value
   write_bigendian_word(out, 255); // max pixel value
   write_zeros(out, 492); // zero padding to make header 512 bytes
   write_channel(out, img->r, img->m*img->n, exposure);
   write_channel(out, img->g, img->m*img->n, exposure);
   write_channel(out, img->b, img->m*img->n, exposure);
   // write_channel(out, img->a, img->m*img->n, 1); uncomment and change to 4 channels if you want to composite on top of a background image
}

// Fragmentation ///////////////////////////////////////////////////////////////////////////////

FragmentList *new_fragment_list(int nmax, FragmentList *next);
void prepend_to_fragment_list(FragmentList **fl, float r, float g, float b, float a, float z);

FragmentList *new_fragment_list(int nmax, FragmentList *next)
{
   FragmentList *fl=malloc(sizeof(FragmentList));
   fl->n=0;
   fl->nmax=nmax;
   fl->rgbaz=(float*)malloc(5*fl->nmax*sizeof(float));
   fl->next=next;
   return fl;
}

void prepend_to_fragment_list(FragmentList **fl, float r, float g, float b, float a, float z)
{
   int n;
   if(*fl==0)
      *fl=new_fragment_list(3, 0);
   else if ((**fl).n==(**fl).nmax)
      *fl=new_fragment_list(2*(**fl).nmax, *fl);
   n=(**fl).n;
   (**fl).rgbaz[5*n]=r;
   (**fl).rgbaz[5*n+1]=g;
   (**fl).rgbaz[5*n+2]=b;
   (**fl).rgbaz[5*n+3]=a;
   (**fl).rgbaz[5*n+4]=z;
   ++(**fl).n;
}

void transform(const float A[4][4], float x[3], float *camx, float *camy, float *camz, float *camw)
{
   *camx=A[0][0]*x[0]+A[0][1]*x[1]+A[0][2]*x[2]+A[0][3];
   *camy=A[1][0]*x[0]+A[1][1]*x[1]+A[1][2]*x[2]+A[1][3];
   *camz=A[2][0]*x[0]+A[2][1]*x[1]+A[2][2]*x[2]+A[2][3];
   *camw=A[3][0]*x[0]+A[3][1]*x[1]+A[3][2]*x[2]+A[3][3];
}

float approx_camera_scale(const float A[4][4], const float x[3], float camx, float camy, float camw)
{
   float b00=A[0][0]*camw-camx*A[3][0], b01=A[0][1]*camw-camx*A[3][1], b02=A[0][2]*camw-camx*A[3][2],
         b10=A[1][0]*camw-camy*A[3][0], b11=A[1][1]*camw-camy*A[3][1], b12=A[1][2]*camw-camy*A[3][2];
   float c00=b00*b00+b01*b01+b02*b02, c10=b10*b00+b11*b01+b12*b02, c11=b10*b10+b11*b11+b12*b12;
   float lambda_max=0.5*(c00+c11+sqrt(sqr(c00-c11)+4*c10*c10));
   return sqrt(lambda_max)/(camw*camw);
}

// smoothly goes from 1 at x=0 to 0 at x=1
inline float kernel(float x)
{
   if(x<0) return 1;
   else if(x>1) return 0;
   else return 1+x*x*(-3+2*x);
}

void fragment_particles(const Particles *pts, const ViewSettings *view, FragmentBuffer *fb)
{
   int p, i, j, istart, iend, jstart, jend;
   float camx, camy, camz, camw, camr;
   float imgx, imgy, imgr;
   float r, g, b, a, s, k;

   for(p=0; p<pts->n; ++p){
      // transform particle to camera space
      transform(view->transform, pts->x+3*p, &camx, &camy, &camz, &camw);
      if(fabs(camz)>=fabs(camw)) continue; // particle not visible by camera
      // also transform particle radius into view radius
      camr=pts->radius[p]*approx_camera_scale(view->transform, pts->x+3*p, camx, camy, camw);
      // normalize homogenous coordinates
      camx/=camw;
      camy/=camw;
      camz/=camw;
      // transform camera stuff to screen stuff
      imgx=0.5*(camx+view->camera_x_max)*fb->n;
      imgy=0.5*(camy+1)*fb->n;
      imgr=0.5*camr*fb->n;
      // antialias by spreading out if too small
      if(imgr<1){
         s=sqr(imgr);
         imgr=1;
      }else
         s=1;
      // figure out fragment ranges and frustum cull
      istart=(int)floor(imgx-imgr);
      if(istart<0) istart=0; else if(istart>=fb->m) continue;
      iend=(int)ceil(imgx+imgr);
      if(iend<0) continue; else if(iend>=fb->m) iend=fb->m-1;
      jstart=(int)floor(imgy-imgr);
      if(jstart<0) jstart=0; else if(jstart>=fb->n) continue;
      jend=(int)ceil(imgy+imgr);
      if(jend<0) continue; else if(jend>=fb->n) jend=fb->n-1;
      // scan convert
      r=pts->rgb[3*p];
      g=pts->rgb[3*p+1];
      b=pts->rgb[3*p+2];
      a=s*pts->opacity[p];
      for(i=istart; i<=iend; ++i){
         for(j=jstart; j<=jend; ++j){
            k=kernel(sqrt(sqr(i+0.5-imgx)+sqr(j+0.5-imgy))/imgr);
            prepend_to_fragment_list(fb->fragments+i+fb->m*j, r, g, b, k*a, camz);  // don't premultiply alpha!
         }
      }
      if(p%100==0){
         fputc(p%1000 ? '+' : '#',stdout); // progress indicator
         fflush(stdout);
      }
   }
   fputc('\n', stdout);
}

// Compositing ///////////////////////////////////////////////////////////////////////////////////////////

int compare_rgbaz(const void *f1, const void *f2)
{
   float z1=((float*)f1)[4], z2=((float*)f2)[4];
   if(z1<z2) return -1;
   else if (z1>z2) return 1;
   else return 0;
}

void composite_pixel(FragmentList **fl, float *r, float *g, float *b, float *a, float *z)
{
   int i, num_frag=0;
   FragmentList *p;
   float *rgbaz, alpha;

   *r=0; *g=0; *b=0; *a=0; *z=1; // default

   // copy fragment list to one array
   for(p=*fl; p; p=p->next)
      num_frag+=p->n;
   if(num_frag==0)
      return;
   rgbaz=malloc(5*num_frag*sizeof(float));
   num_frag=0;
   for(p=*fl; p; p=p->next){
      memcpy(rgbaz+5*num_frag, p->rgbaz, 5*p->n*sizeof(float));
      num_frag+=p->n;
   }
   // sort it into increasing z
   qsort(rgbaz, num_frag, 5*sizeof(float), compare_rgbaz);
   // and composite
   for(i=num_frag-1; i>=0; --i){
      alpha=rgbaz[5*i+3];
      *r=alpha*rgbaz[5*i]+(1-alpha)*(*r);
      *g=alpha*rgbaz[5*i+1]+(1-alpha)*(*g);
      *b=alpha*rgbaz[5*i+2]+(1-alpha)*(*b);
      *a=alpha+(1-alpha)*(*a);
   }
   *z=rgbaz[4]; // pixel z value is frontmost z
   free(rgbaz);
}

void composite_fragments(FragmentBuffer *fb, ImageRGBAZ *img)
{
   int i;
   for(i=0; i<fb->m*fb->n; ++i){
      composite_pixel(fb->fragments+i, img->r+i, img->g+i, img->b+i, img->a+i, img->z+i);
      if(i%10000==0){
         fputc(i%100000 ? '+' : '#', stdout); // progress indicator
         fflush(stdout);
      }
   }
   fputc('\n', stdout);
}

